/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __FLEXIO_DEV_H__
#define __FLEXIO_DEV_H__

#include <stdint.h>
#include <common/flexio_common.h>

#define      __unused        __attribute__((__unused__))
#define __dpa_global__ __attribute__((annotate("__dpa_global__")))
#define __dpa_rpc__ __attribute__((annotate("__dpa_rpc__")))

/* Entry point start and end macros */
#define __FLEXIO_ENTRY_POINT_START _Pragma("clang section text=\".entry_point\"")
#define __FLEXIO_ENTRY_POINT_END   _Pragma("clang section text=\"\"")

/* Return status of flexio_dev API functions */
typedef enum {
	FLEXIO_DEV_STATUS_SUCCESS	= 0,
	FLEXIO_DEV_STATUS_FAILED	= 1,
} flexio_dev_status_t;

enum cq_ce_mode {
	MLX5_CTRL_SEG_CE_CQE_ON_CQE_ERROR        = 0x0,
	MLX5_CTRL_SEG_CE_CQE_ON_FIRST_CQE_ERROR  = 0x1,
	MLX5_CTRL_SEG_CE_CQE_ALWAYS              = 0x2,
	MLX5_CTRL_SEG_CE_CQE_AND_EQE             = 0x3,
};

struct flexio_dev_thread_ctx;

/* Function types for NET and RPC handlers */

/**
 * @brief RPC handler callback function type.
 *
 * Defines an RPC handler callback function.
 *
 * @param[in] arg - argument of the RPC function.
 *
 * @return uint64_t - result of the RPC function.
 */
typedef uint64_t (flexio_dev_rpc_handler_t)(uint64_t arg);

/**
* @brief Unpack the arguments and call the user function
*
* This callback function is used at runtime to unpack the
* arguments from the call on Host and then call the function on DPA.
* This function is called internally from flexio dev.
*
* @param argbuf [in] - Argument buffer that was written by Host
* @param func [in] - Function pointer to user function
*
* @return uint64_t - result of the RPC function.
*/
typedef uint64_t (flexio_dev_arg_unpack_func_t) (void *argbuf, void *func);

/**
 * @brief event handler callback function type.
 *
 * Defines an event handler callback function.
 * On handler function end, need to call flexio_dev_finish() instead of a regular return statement,
 * in order to properly release resources back to the OS.
 *
 * @param[in] thread_arg - an argument for the executing thread.
 *
 * @return void.
 */
typedef void (flexio_dev_event_handler_t)(uint64_t thread_arg);

/**
 * @brief poll thread handler callback function type.
 *
 * Defines a poll thread handler callback function.
 *
 * @param[in] thread_arg - an argument for the executing thread.
 *
 * @return void.
 */
typedef void (flexio_dev_poll_thread_t)(uint64_t thread_arg);

struct spinlock_s {
	uint32_t locked;
};

/**
 * @brief Initialize a spinlock mechanism.
 *
 * Initialize a spinlock mechanism, must be called before use.
 *
 * @param[in]  A pointer to spinlock_s structure.
 *
 * @return void.
 */
#define spin_init(lock)\
	__atomic_exchange_n(&((lock)->locked), 0, __ATOMIC_SEQ_CST)

/**
 * @brief Lock a spinlock mechanism.
 *
 * Lock a spinlock mechanism.
 *
 * @param[in]  A pointer to spinlock_s structure.
 *
 * @return void.
 */
#define spin_lock(lock)         \
do {                            \
	uint32_t __old_val;     \
	do {                    \
		__old_val = 0;  \
	} while (!(__atomic_compare_exchange_n(&((lock)->locked), &__old_val, 1, 0,  \
		__ATOMIC_SEQ_CST, __ATOMIC_SEQ_CST)));                               \
} while (0)

/**
 * @brief Unlock a spinlock mechanism.
 *
 Unlock a spinlock mechanism.
 *
 * @param[in]  A pointer to spinlock_s structure.
 *
 * @return void.
 */
#define spin_unlock(lock)\
	__atomic_exchange_n(&((lock)->locked), 0, __ATOMIC_SEQ_CST)

/**
 * @brief Request thread context.
 *
 * This function requests the thread context. Should be called
 * for every start of thread.
 *
 * @param[in]  dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 *
 * @return 0 on success negative value on failure.
 */
int flexio_dev_get_thread_ctx(struct flexio_dev_thread_ctx **dtctx);

/**
 * @brief Exit from a thread, leave process active
 *
 * This function releases resources back to OS.
 * For the next DUAR the thread will restart from the beginning.
 *
 * @param[in]  void.
 *
 * @return Function does not return
 */
void flexio_dev_reschedule(void);

/**
 * @brief Exit flexio process (no errors)
 *
 * This function releases resources back to OS and returns '0x40' in dpa_process_status
 *
 * @param[in]  void.
 *
 * @return Function does not return
 */
void flexio_dev_finish(void);

/**
 * @brief Put a string to printing queue.
 *
 * This function puts a string to host printing queue.
 * This queue has been serviced by host application.
 * Would have no effect, if the host application didn't configure printing environment.
 * In order to initalize/configure printing envoirment -
 * On HOST side - after flexio_process_create, flexio_print_init should be called.
 * On DEV side - before using flexio_dev_puts, the thread context is needed, therefore
 * flexio_dev_get_thread_ctx should be called before.
 *
 * @param[in]  dtctx - A pointer to a pointer of flexio_dev_thread_ctx structure.
 * @param[in]  str   - A pointer to string
 *
 * @return length of printed string.
 */
int flexio_dev_puts(struct flexio_dev_thread_ctx *dtctx, char *str);

/**
 * @brief Config thread outbox object.
 *
 * This function updates the thread outbox object of the given thread
 * context.
 *
 * @param[in]  dtctx - A pointer to flexio_dev_thread_ctx structure.
 * @param[in]  outbox_config_id - The outbox object config id.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_outbox_config(struct flexio_dev_thread_ctx *dtctx,
					     uint16_t outbox_config_id);

/**
 * @brief Config thread window object.
 *
 * This function updates the thread window object of the given thread
 * context.
 *
 * @param[in]  dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in]  window_config_id - The window object id.
 * @param[in]  mkey - mkey object.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_window_config(struct flexio_dev_thread_ctx *dtctx,
					     uint16_t window_config_id, uint32_t mkey);

/**
 * @brief Config thread window mkey object.
 *
 * This function updates the thread window mkey object of the given thread
 * context.
 *
 * @param[in]  dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in]  mkey - mkey object.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_window_mkey_config(struct flexio_dev_thread_ctx *dtctx,
						  uint32_t mkey);

/**
 * @brief Generate device address from host allocated memory.
 *
 * This function generates a memory address to be used by device to access host side memory,
 * according to already create window object. from a host allocated address.
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in] haddr - Host allocated address.
 * @param[out] daddr - A pointer to write the device generated matching address.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_window_ptr_acquire(struct flexio_dev_thread_ctx *dtctx,
						  uint64_t haddr, flexio_uintptr_t **daddr);

/**
 * @brief Copy a buffer from host memory to device memory.
 *
 * This function copies specified number of bytes from host memory to device memory.
 * UNSUPPORTED at this time.
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in] daddr - A pointer to the device memory buffer.
 * @param[in] haddr - A pointer to the host memory allocated buffer.
 * @param[in] size - Number of bytes to copy.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_window_copy_from_host(struct flexio_dev_thread_ctx *dtctx,
						     void *daddr, uint64_t haddr, uint32_t size);

/**
 * @brief Copy a buffer from device memory to host memory.
 *
 * This function copies specified number of bytes from device memory to host memory.
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.
 * @param[in] haddr - A pointer to the host memory allocated buffer.
 * @param[in] daddr - A pointer to the device memory buffer.
 * @param[in] size - Number of bytes to copy.
 *
 * @return flexio_dev_status_t.
 */
flexio_dev_status_t flexio_dev_window_copy_to_host(struct flexio_dev_thread_ctx *dtctx,
						   uint64_t haddr, const void *daddr,
						   uint32_t size);

/**
 * @brief Get thread ID from thread context
 *
 * This function queries a thread context for its thread ID (from thread metadata).
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.

 * @return thread ID value.
 */
uint32_t flexio_dev_get_thread_id(struct flexio_dev_thread_ctx *dtctx);

/**
 * @brief Get thread local storage address from thread context
 *
 * This function queries a thread context for its thread local storage (from thread metadata).
 *
 * @param[in] dtctx - A pointer to a flexio_dev_thread_ctx structure.

 * @return thread local storage value.
 */
flexio_uintptr_t flexio_dev_get_thread_local_storage(struct flexio_dev_thread_ctx *dtctx);

/**
 * @brief Same as a regular printf but with protection from simultaneous print
 * from different threads.
 *
 * @params[in] - same as for regular printf.

 * @return - same as from regular printf.
 */
int flexio_dev_print(const char *format, ...) __attribute__ ((format(printf, 1, 2)));

#endif
