/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __FLEXIO_DEV_QUEUE_TYPES_H__
#define __FLEXIO_DEV_QUEUE_TYPES_H__

#include <libflexio-dev/flexio_dev.h>
#include <stdint.h>

/* SQ depth (log_sq_depth) is measured in WQEBBs, each one is 64B.
 * We have to understand difference between wqe_idx and seg_idx.
 * For example wqe with index 5 built from 4 segments with indexes 20, 21, 22 and 23
 */
#define LOG_SQE_NUM_SEGS 2

typedef uint64_t __be64;
typedef uint32_t __be32;
typedef uint16_t __be16;

/* This stuff based on content of mlx5/device.h file
 * from MFT driver
 */
struct flexio_dev_eqe {
	uint8_t rsvd00;       /* 00h */
	uint8_t type;         /* 01h */
	uint8_t rsvd02;       /* 02h */
	uint8_t sub_type;     /* 03h */

	uint8_t rsvd4[28];    /* 04h...1fh */
	struct {
		__be32 rsvd00[6];  /* 00h..17h */
		__be32 cqn;       /* 18h 24 lsb */
	} event_data;         /* 20h */

	__be16 rsvd3c;        /* 3ch */
	uint8_t signature;    /* 3eh */
	uint8_t owner;        /* 3fh */
} __attribute__((packed, aligned(8)));

struct flexio_dev_cqe64 {
	__be32 rsvd0[8];
	__be32 srqn_uidx;
	__be32 rsvd36[2];
	__be32 byte_cnt;
	__be32 rsvd48[2];
	__be32 qpn;
	__be16 wqe_counter;
	uint8_t signature;
	uint8_t op_own;
} __attribute__((packed, aligned(8)));

struct flexio_dev_wqe_data_seg {
	__be32 byte_count;
	__be32 lkey;
	__be64 addr;
} __attribute__((packed, aligned(8)));

struct flexio_dev_wqe_ctrl_seg {
	__be32 idx_opcode;
	__be32 qpn_ds;
	__be32 signature_fm_ce_se;
	__be32 general_id;
} __attribute__((packed, aligned(8)));

/* PRM section 8.9.4.2, Table 49 Send WQE Construction Summary */
typedef enum {
	FLEXIO_ETH_SEG_L4CS 			= 0x8000,
	FLEXIO_ETH_SEG_L3CS 			= 0x4000,
	FLEXIO_ETH_SEG_L4CS_INNER		= 0x2000,
	FLEXIO_ETH_SEG_L3CS_INNER		= 0x1000,
	FLEXIO_ETH_SEG_TRAILER_ALIGN		= 0x0200,
	FLEXIO_ETH_SEG_SWP_OUTER_L4_TYPE	= 0x0040,
	FLEXIO_ETH_SEG_SWP_OUTER_L3_TYPE	= 0x0020,
	FLEXIO_ETH_SEG_SWP_INNER_L4_TYPE	= 0x0002,
	FLEXIO_ETH_SEG_SWP_INNER_L3_TYPE	= 0x0001,
} flexio_dev_wqe_eth_seg_cs_swp_flags_t;

struct flexio_dev_wqe_eth_seg {
	__be32 rsvd0;
	__be16 cs_swp_flags;
	__be16 mss;
	__be32 rsvd2;
	__be16 inline_hdr_bsz;
	uint8_t inline_hdrs[2];
}  __attribute__((packed, aligned(8)));

struct flexio_dev_wqe_inline_data_seg {
	uint8_t inline_data[16];
}  __attribute__((packed, aligned(8)));

struct flexio_dev_wqe_rdma_seg {
	__be64 raddr;
	__be32 rkey;
	__be32 rsvd0;
}  __attribute__((packed, aligned(8)));

struct flexio_dev_wqe_atomic_seg {
	__be64 swap_or_add_data;
	__be64 compare_data;
}  __attribute__((packed, aligned(8)));

union flexio_dev_sqe_seg {
	struct flexio_dev_wqe_ctrl_seg ctrl;
	struct flexio_dev_wqe_eth_seg eth;
	struct flexio_dev_wqe_inline_data_seg inline_data;
	struct flexio_dev_wqe_rdma_seg rdma;
	struct flexio_dev_wqe_data_seg data;
	struct flexio_dev_wqe_atomic_seg atomic;
};

#endif
