/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _FLEXIO_SDK_H_
#define _FLEXIO_SDK_H_

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PHY_CORES_NUM 16
#define HARTS_PER_CORE 16
#define HARTS_LIMIT (PHY_CORES_NUM * HARTS_PER_CORE)
#define OS_RESERVED_HARTS 2
#define USER_HARTS_LIMIT (HARTS_LIMIT - OS_RESERVED_HARTS)
#define HART_BITMASK_LEN (HARTS_LIMIT / 8)

typedef uint64_t flexio_uintptr_t;

enum {
	FLEXIO_CQ_ELEMENT_TYPE_DPA_THREAD	= 0x0,
	FLEXIO_CQ_ELEMENT_TYPE_EXT_EQ		= 0x2,
	/* value 0x1 reserved */
	FLEXIO_CQ_ELEMENT_TYPE_NON_DPA_CQ	= 0x8,
	FLEXIO_CQ_ELEMENT_TYPE_DPA_MSIX_EMULATED_CQ = 0x9,
};

/* flexio SDK types and structs */

typedef enum {
	FLEXIO_STATUS_SUCCESS	= 0,
	FLEXIO_STATUS_FAILED	= 1,
	FLEXIO_STATUS_TIMEOUT	= 2,
	FLEXIO_STATUS_FATAL_ERR = 3,
} flexio_status;

typedef enum {
	FLEXIO_QPC_ST_RC            = 0x0,
	FLEXIO_QPC_ST_UC            = 0x1,
	FLEXIO_QPC_ST_UD            = 0x2,
	FLEXIO_QPC_ST_XRC           = 0x3,
	FLEXIO_QPC_ST_IBL2          = 0x4,
	FLEXIO_QPC_ST_DCI           = 0x5,
	FLEXIO_QPC_ST_QP0           = 0x7,
	FLEXIO_QPC_ST_QP1           = 0x8,
	FLEXIO_QPC_ST_RAW_DATAGRAM  = 0x9,
	FLEXIO_QPC_ST_REG_UMR       = 0xc,
	FLEXIO_QPC_ST_DC_CNAK       = 0x10,
} flexio_qp_transport_type;

typedef enum {
	FLEXIO_CQE_COMP_NONE	= 0x0,
	FLEXIO_CQE_COMP_BASIC	= 0x1,
	FLEXIO_CQE_COMP_ENH	= 0x2,
} flexio_cqe_comp_type;

typedef enum {
	FLEXIO_CQ_PERIOD_MODE_EVENT	= 0x0,
	FLEXIO_CQ_PERIOD_MODE_CQE	= 0x1,
} flexio_cq_period_mode_t;

typedef void (flexio_func_t) (void);

struct flexio_process;
struct flexio_window;
struct flexio_outbox;
struct flexio_event_handler;
struct flexio_thread;
struct flexio_cq;
struct flexio_rq;
struct flexio_sq;
struct flexio_qp;
struct flexio_mkey;
struct flexio_uar;
struct flexio_msix;
struct flexio_emu_device_msix_map;
struct flexio_eq_msix_list;
struct flexio_emu_db_to_cq_ctx;
struct flexio_transport_domain;
struct flexio_app;
struct flexio_func;

enum flexio_memtype {
	FLEXIO_MEMTYPE_DPA = 0, /* Usage of zero value for DPA memory works as a default type */
	FLEXIO_MEMTYPE_HOST = 1,
};

struct flexio_qmem {
	enum flexio_memtype memtype;
	union {
		flexio_uintptr_t daddr;
		void *haddr;
	};
	uint32_t umem_id;
};

struct flexio_cq_attr {
	uint8_t log_cq_depth;
	uint8_t element_type;
	union {
		uint32_t *eqn;
		struct flexio_thread *thread;
		struct flexio_msix* msix;
	};
	uint32_t uar_id;
	void *uar_base_addr;
	flexio_cqe_comp_type cqe_comp_type;
	flexio_cq_period_mode_t cq_period_mode;
	uint16_t cq_period;
	uint16_t cq_max_count;
	bool cc;	/* Indicates CQ entries are collapsed to first CQE */

	/* memory allocated for queue */
	struct flexio_qmem cq_ring_qmem;
	flexio_uintptr_t cq_dbr_daddr;
};

struct flexio_wq_sq_attr {
	uint8_t allow_multi_pkt_send_wqe;
};

struct flexio_wq_attr {
	uint8_t log_wq_depth;
	uint32_t uar_id;
	uint32_t user_index;
	struct ibv_pd *pd;
	union {
		struct flexio_wq_sq_attr sq;
	};

	/* Memory allocated for queue */
	struct flexio_qmem wq_ring_qmem;
	flexio_uintptr_t wq_dbr_daddr;
};

struct flexio_process_attr {
	struct ibv_pd *pd;
};

enum flexio_affinity_type {
	FLEXIO_AFFINITY_NONE,
	FLEXIO_AFFINITY_STRICT,
	FLEXIO_AFFINITY_GROUP,
};

typedef struct flexio_affinity {
	enum flexio_affinity_type type;
	/* ID of the chosen resource (HART / DPA HART group). Reserved if affinity is no affinity */
	uint32_t id;
} flexio_affinity;

struct flexio_event_handler_attr {
	flexio_func_t *host_stub_func;
	uint64_t arg;
	flexio_uintptr_t thread_local_storage_daddr;
	flexio_affinity affinity;
};

enum flexio_qp_op_types {
	FLEXIO_QP_WR_RDMA_WRITE = 0x4,
	FLEXIO_QP_WR_RDMA_READ = 0x8,
	FLEXIO_QP_WR_ATOMIC_CMP_AND_SWAP = 0x10,
};

struct flexio_qp_attr_opt_param_mask {
	bool qp_access_mask;
	bool min_rnr_nak_timer;
};

enum flexio_qp_qpc_mtu {
	FLEXIO_QP_QPC_MTU_BYTES_256 = 0x1,
	FLEXIO_QP_QPC_MTU_BYTES_512 = 0x2,
	FLEXIO_QP_QPC_MTU_BYTES_1K  = 0x3,
	FLEXIO_QP_QPC_MTU_BYTES_2K  = 0x4,
	FLEXIO_QP_QPC_MTU_BYTES_4K  = 0x5,
};

typedef enum flexio_qp_state {
	FLEXIO_QP_STATE_RST,
	FLEXIO_QP_STATE_INIT,
	FLEXIO_QP_STATE_RTR,
	FLEXIO_QP_STATE_RTS,
} flexio_qp_state;

struct flexio_qp_attr {
	uint32_t transport_type;
	uint32_t uar_id;
	uint32_t user_index;
	int qp_access_mask;	/* Bitmask of enum ibv_access_flags.
				   Expected values: IBV_ACCESS_REMOTE_WRITE, IBV_ACCESS_REMOTE_READ,
				   IBV_ACCESS_REMOTE_ATOMIC, IBV_ACCESS_LOCAL_WRITE. */
	int ops_flag;	/* Bitmask of enum flexio_qp_op_types, used to calculate WQE sizes */
	struct ibv_pd *pd;

	int log_sq_depth;
	int no_sq;
	uint32_t sq_cqn;

	int log_rq_depth;
	/* BUG FIXME:  MLX5_QPC_RQ_TYPE_* has not been visible to user */
	int rq_type; /* MLX5_QPC_RQ_TYPE_* */
	uint32_t rq_cqn;

	struct flexio_qmem qp_wq_buff_qmem;
	flexio_uintptr_t qp_dbr_daddr;

	/* RC QP modify attributes */
	flexio_qp_state next_state;
	uint32_t remote_qp_num;
	uint8_t gid_table_index;
	uint8_t fl; /* Force Loopback */
	uint8_t *dest_mac;
	union ibv_gid rgid_or_rip;
	uint16_t rlid;
	uint32_t min_rnr_nak_timer;
	enum flexio_qp_qpc_mtu path_mtu;
	uint8_t retry_count;
	uint8_t vhca_port_num;
	uint32_t next_rcv_psn;
	uint32_t next_send_psn;
	uint16_t udp_sport;
	uint8_t grh;
	uint8_t log_rra_max;
};

struct flexio_mkey_attr {
	struct ibv_pd *pd;
	flexio_uintptr_t daddr;
	size_t len;
	int access;
};

/* Flex IO API Functions */

/**
 * @brief Allocates a buffer on Flex IO heap memory.
 *
 * This function allocates a buffer with the requested size on the Flex IO heap memory.
 * On success - sets dest_daddr_p to the start address of the allocated buffer.
 * On Failure - sets dest_daddr_p to 0x0.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  buff_bsize - The size of the buffer to allocate.
 * @param[out] dest_daddr_p - A pointer to the Flex IO address, where the buffer was allocated.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_alloc(struct flexio_process *process, size_t buff_bsize,
				   flexio_uintptr_t *dest_daddr_p);

/**
 * @brief Deallocates Flex IO heap memory buffer.
 *
 * This function frees Flex IO heap memory buffer by address.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] daddr - A pointer to an address of allocated memory on the Flex IO heap.
 *		      Zero value is valid argument.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_free(struct flexio_process *process, flexio_uintptr_t daddr);

/**
 * @brief Copy from host memory to a pre-allocted Flex IO heap memory buffer.
 *
 * This function copies data from a buffer on the host memory to a buffer on the Flex IO heap
 * memory.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] src_haddr - An address of the buffer on the host memory.
 * @param[in] buff_bsize - The size of the buffer to copy.
 * @param[in] dest_daddr - Flex IO heap memory buffer address to copy to.
 *
 * @return flexio status value.
 */
flexio_status flexio_host2dev_memcpy(struct flexio_process *process, void *src_haddr,
				     size_t buff_bsize, flexio_uintptr_t dest_daddr);

/**
 * @brief Sets DPA heap memory buffer to a given value.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] value - A value to set the DPA heap memory buffer to.
 * @param[in] buff_bsize - The size of the Flex IO heap memory buffer.
 * @param[in] dest_daddr - Flex IO heap memory buffer address to set.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_memset(struct flexio_process *process, int value,
				    size_t buff_bsize, flexio_uintptr_t dest_daddr);

/**
 * @brief Copy from host memory to Flex IO heap memory buffer.
 *
 * This function copies data from a buffer on the host memory to the Flex IO memory.
 * The function allocates memory on the device heap which dest_address points to.
 * It is the caller responsibility to deallocate this memory when it is no longer used.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  src_haddr - An address of the buffer on the host memory.
 * @param[in]  buff_bsize - The size of the buffer to copy.
 * @param[out] dest_daddr_p - A pointer to the Flex IO address, where the buffer was copied to.
 *
 * @return flexio status value.
 */
flexio_status flexio_copy_from_host(struct flexio_process *process, void *src_haddr,
				    size_t buff_bsize, flexio_uintptr_t *dest_daddr_p);

/**
 * @brief Creates a Flex IO CQ.
 *
 * This function creates a Flex IO CQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * 			 If NULL - process' will be used.
 * @param[in]  fattr - A pointer to the CQ attributes struct.
 * @param[out] cq - A pointer to the created CQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       const struct flexio_cq_attr *fattr, struct flexio_cq **cq);

/**
 * @brief Destroys a Flex IO CQ.
 *
 * This function destroys a Flex IO CQ.
 *
 * @param[in] cq - A pointer to a CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_destroy(struct flexio_cq *cq);

/**
 * @brief Queries a Flex IO CQ moderation configuration.
 *
 * @param[in] cq - A pointer to a CQ context.
 * @param[out] max_count - A pointer to the CQ moderation max count value.
 * @param[out] period - A pointer to the CQ moderation period value.
 * @param[out] mode - A pointer to the CQ moderation mode value.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_query_moderation(struct flexio_cq *cq,
					 uint16_t *max_count,
					 uint16_t *period,
					 uint16_t *mode);

/**
 * @brief Modifies a Flex IO CQ moderation configuration.
 *
 * @param[in] cq - A pointer to a CQ context.
 * @param[out] max_count - CQ moderation max count value.
 * @param[out] period - CQ moderation period value.
 * @param[out] mode - CQ moderation mode value.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_modify_moderation(struct flexio_cq *cq,
					  uint16_t max_count,
					  uint16_t period,
					  uint16_t mode);

/**
 * @brief Creates an Mkey to the process device UMEM
 *
 * This function creates an MKey over the provided PD for the provided process device
 * UMEM.
 * The mkey_id will point to the field in the containing flexio_mkey object.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  fattr - A pointer to a Flex IO MKey attribute struct.
 * @param[out] mkey - A pointer to a pointer to the created MKey struct.
 *
 * @return flexio status value.
 */
flexio_status flexio_device_mkey_create(struct flexio_process *process,
					struct flexio_mkey_attr *fattr, struct flexio_mkey **mkey);

/**
 * @brief destroys an MKey object containing the given ID
 *
 * This function destroys an Mkey object containing the given ID.
 *
 * @param[in] mkey - A pointer to the Flex IO MKey to destroy. NULL is a valid value.
 *
 * @return flexio status value.
 */
flexio_status flexio_device_mkey_destroy(struct flexio_mkey *mkey);

/**
 * @brief Creates a Flex IO UAR object
 *
 * This function creates a Flex IO UAR object from a DevX UAR object.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  devx_uar - A pointer to a DevX UAR struct.
 * @param[out] flexio_uar - A pointer to a pointer to the created Flex IO UAR struct.
 *
 * @return flexio status value.
 */
flexio_status flexio_uar_create(struct flexio_process *process, struct mlx5dv_devx_uar *devx_uar,
				struct flexio_uar **flexio_uar);

/**
 * @brief destroys a Flex IO UAR object
 *
 * This function destroys a Flex IO UAR object.
 *
 * @param[in] uar - A pointer to the Flex IO UAR to destroy.
 *
 * @return flexio status value.
 */
flexio_status flexio_uar_destroy(struct flexio_uar *uar);

/**
 * @brief Creates a Flex IO event handler.
 *
 * This function creates a Flex IO event handler for an existing Flex IO process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  fattr - A pointer to the event handler attributes struct.
 * @param[in]  window_id - Default window ID (or NULL for no default window).
 * @param[in]  outbox_id - Default outbox ID (or NULL for no default outbox).
 * @param[out] event_handler_ptr - A pointer to the created event handler context pointer.
 *
 * @return flexio status value.
 */
flexio_status
flexio_event_handler_create(struct flexio_process *process, struct flexio_event_handler_attr *fattr,
			    const struct flexio_window *window, const struct flexio_outbox *outbox,
			    struct flexio_event_handler **event_handler_ptr);

/**
 * @brief Destroys a Flex IO event handler.
 *
 * This function destroys a Flex IO event handler.
 *
 * @param[in] event_handler - A pointer to an event handler context.
 *
 * @return flexio status value.
 */
flexio_status flexio_event_handler_destroy(struct flexio_event_handler *event_handler);

/**
 * @brief Run a Flex IO event handler.
 *
 * This function makes a Flex IO event handler start running.
 *
 * @param[in] event_handler - A pointer to an event handler context.
 * @param[in] user_arg - A 64 bit argument for the event handler's thread.
 *
 * @return flexio status value.
 */
flexio_status
flexio_event_handler_run(struct flexio_event_handler *event_handler, uint64_t user_arg);

/**
 * @brief Calls a Flex IO process.
 *
 * @param[in] process - A pointer to the Flex IO process to run.
 * @param[in] host_func - The host stub function that is used by the application to
 *                        reference the device function.
 * @param[out] func_ret - A pointer to the ELF function return value.
 * @param[in] args - Arguments (var argos).
 *
 * @return flexio status value.
 */
flexio_status flexio_process_call(struct flexio_process *process, flexio_func_t *host_func,
				  uint64_t *func_ret, ... /* args for the user function */);

/**
 * @brief Create a new Flex IO process.
 *
 * This function creates a new Flex IO process with requested image.
 *
 * @param[in]  ibv_ctx - A pointer to a device context.
 * @param[in]  app - Device side application handle.
 * @param[in]  process_attr - Optional, process attributes for create. Can be NULL.
 * @param[out] process_ptr - A pointer to the created process pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_process_create(struct ibv_context *ibv_ctx, struct flexio_app *app,
				    const struct flexio_process_attr *process_attr,
				    struct flexio_process **process_ptr);

/**
 * @brief Set the Flexio process error handler.
 *
 * This function sets the Flexio process error handler. The error handler must be set after
 * the process is created, and before the first thread is created.
 *
 * @param[in] process - A pointer to a process
 * @param[in] handler - a C string of the function name in the dpa elf file image passed to
 *                      flexio_process_create
 *
 * @return flexio status value.
 */
flexio_status flexio_process_error_handler_set(struct flexio_process *process, const char *handler);

/**
 * @brief Destroys a new Flex IO process.
 *
 * This function destroys a new Flex IO process.
 *
 * @param[in] process - A pointer to a process.
 *
 * @return flexio status value.
 */
flexio_status flexio_process_destroy(struct flexio_process *process);

/**
 * @brief Creates a Flex IO RQ.
 *
 * This function creates a Flex IO RQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * 			 If NULL - process' will be used.
 * @param[in]  cq_num - A CQ number.
 * @param[in]  fattr - A pointer to the RQ attributes struct.
 * @param[out] rq - A pointer to the created RQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_rq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       uint32_t cq_num, const struct flexio_wq_attr *fattr,
			       struct flexio_rq **flexio_rq_ptr);

/**
 * @brief Destroys a Flex IO RQ.
 *
 * This function destroys a Flex IO RQ.
 *
 * @param[in] rq - A pointer to an RQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_rq_destroy(struct flexio_rq *flexio_rq);

/**
 * @brief Creates a Flex IO SQ.
 *
 * This function creates a Flex IO SQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * 			 If NULL - process' will be used.
 * @param[in]  cq_num - A CQ number (can be Flex IO or host CQ).
 * @param[in]  fattr - A pointer to the SQ attributes struct.
 * @param[out] sq - A pointer to the created SQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_sq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       uint32_t cq_num, const struct flexio_wq_attr *fattr,
			       struct flexio_sq **flexio_sq_ptr);

/**
 * @brief Destroys a Flex IO SQ.
 *
 * This function destroys a Flex IO SQ.
 *
 * @param[in] sq - A pointer to an SQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_sq_destroy(struct flexio_sq *flexio_sq);

/**
 * @brief Creates a Flex IO window.
 *
 * This function Creates a Flex IO window for the given process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  pd - A pointer to a protection domain struct to the memory the window should access.
 * @param[out] window - A pointer to the created window context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_window_create(struct flexio_process *process, struct ibv_pd *pd,
				   struct flexio_window **window);

/**
 * @brief Destroys a Flex IO window.
 *
 * This function destroys a Flex IO window.
 *
 * @param[in] window - A pointer to a window context.
 *
 * @return flexio status value.
 */
flexio_status flexio_window_destroy(struct flexio_window *window);

/**
 * @brief Creates a Flex IO outbox.
 *
 * This function Creates a Flex IO outbox for the given process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  other_ctx - An IBV context for creating the PRM object (Different than process's on
 *                         multi GVMI case, NULL or same as process's otherwise).
 * @param[in]  uar - A Flex IO UAR struct created for the outbox's IBV device.
 * @param[out] outbox - A pointer to the created outbox context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_outbox_create(struct flexio_process *process, struct ibv_context *other_ctx,
				   struct flexio_uar *uar, struct flexio_outbox **outbox);

/**
 * @brief Destroys a Flex IO outbox.
 *
 * This function destroys a Flex IO outbox.
 *
 * @param[in] outbox - A pointer to a outbox context.
 *
 * @return flexio status value.
 */
flexio_status flexio_outbox_destroy(struct flexio_outbox *outbox);

/**
 * @brief Create environment to support print from DPA.
 *
 * This function allocates resources to support printing from Flex IO to HOST.
 *
 * Print works in synchronous or asynchronous modes.
 * In case of synchronous mode dedicated thread started to receive data and print it immediately.
 * In case of asynchronous mode all print buffer will be flushed by flexio_print_flush().
 * Buffer can be overrun.
 *
 * This function doesn't have "destroy" pair. All printing infrastructure will be closed
 * and resources will be released from flexio_process_destroy() function.
 *
 * @param[in] process - A pointer to the Flex IO process.
 * @param[in] flexio_uar - A pointer to a Flex IO UAR object created by caller for device side
 * @param[in] data_bsize - size of buffer, used for data transfer from Flex IO to HOST
 * 			  MUST be power of two and be at least 2Kb
 * @param[in] out - file to save data from Flex IO. Use stdout if you want receive data
 * 		    on HOST's console
 * @param[in] is_async - select between sync(recommended)/async modes
 * @param[out] ppthread - A pointer to receive phread ID of created thread. May be NULL if
 * 			  user doesn't need it.
 *
 * @return flexio status value.
 */
flexio_status flexio_print_init(struct flexio_process *process, struct flexio_uar *flexio_uar,
				size_t data_bsize, FILE *out, int is_async, pthread_t *ppthread);

/**
 * @brief Flush print buffer in case of asynchronous print.
 *
 * All data from print buffer will be flushed to file, definded in flexio_print_init().
 *
 * In case of synchronous print this functions does nothing.
 * This function allocates resources to support printing from Flex IO to HOST.
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return flexio status value.
 */
flexio_status flexio_print_flush(struct flexio_process *process);

/**
 * @brief Destroys a flexio print environment.
 *
 * This function destroys and releases all resources, allocated for process printing needs by
 * flexio_print_init().
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return flexio status value.
 */
flexio_status flexio_print_destroy(struct flexio_process *process);
/**
 * @brief Get file descriptor for error handler
 *
 * User should get fd in order to monitor for nonrecoverable errors
 *
 * User can poll all created processes, using select/poll/epoll
 * functions family.
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return - file descriptor.
 */
int flexio_err_handler_fd(struct flexio_process *process);

/**
 * @brief Check if unrecoverable error occurred
 *
 * User can call this function many times. Error status will be reset only after calling
 * flexio_err_display();
 *
 * It is suggested to check error status after every negotiation with DPA and periodically later.
 * Alternative way - to poll file descriptor got from flexio_err_handler_fd() in order to get
 * events about error
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return - nonzero value if error happen.
 */
int flexio_err_status(struct flexio_process *process);

/**
 * @brief Print error data to stderr
 *
 * User should call this function, once he got information, that an unrecoverable error in DPA has
 * happened.
 *
 * Calling this function will reset error status.
 *
 * @param[in] process - A pointer to the Flex IO process.
 */
void flexio_err_display(struct flexio_process *process);

#define FLEXIO_MAX_EQ_TO_MSIX_MAP_PAIRS 64
/**
 * @brief Query mapping of EQs to MSI-X vectors.
 *
 * This function gets emulated devices VHCA ID and return mapping of EQs to MSI-X vectors
 * for this emulated device.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  emualterd_device_vhca_id - An emulated device VHCA ID.
 * @param[in]  max_msix - maximal msix to query, should not exceed FLEXIO_MAX_EQ_TO_MSIX_MAP_PAIRS
 * @param[out] map - A pointer to receive list of EQs to MSI-X vectors mapping.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_map_create(struct flexio_process *process,
						     uint32_t device_emulation_id,
						     uint16_t max_msix,
						     struct flexio_emu_device_msix_map** map);

/**
 * @brief Free list of EQs to MSI-X vectors mapping.
 *
 * This function frees a list of EQs to MSI-X vectors mapping.
 *
 * @param[in] map - A point to list of EQs to MSI-X vectors mapping.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_map_free(struct flexio_emu_device_msix_map* map);

/**
 * @brief Mapping emulated devices doorbell to CQ.
 *
 * This function maps emulated device doorbell to Flex IO CQ.
 * A doorbell from the emulated device queue will create CQE on the specified CQ.
 *
 * @param[in]  ibv_ctx - A pointer to a device context.
 * @param[in]  emu_dev_vhca_id - An emulated device VHCA ID.
 * @param[in]  emu_dev_queue_id - A queue of the emulated device.
 * @param[in]  cq - A pointer to the Flex IO CQ.
 * @param[out] emu_db_to_cq_ctx - A pointer to the created emulated device doorbell to CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_emu_db_to_cq_map(struct ibv_context *ibv_ctx, uint32_t emu_dev_vhca_id,
				      uint32_t emu_dev_queue_id, struct flexio_cq *cq,
				      struct flexio_emu_db_to_cq_ctx** emu_db_to_cq_ctx);

/**
 * @brief Unmapping emulated devices doorbell to CQ.
 *
 * This function unmaps emulated device queue doorbell to Flex IO CQ.
 *
 * @param[in] emu_device_to_cq_ctx - A pointer to emulated device doorbell to CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_emu_db_to_cq_unmap(struct flexio_emu_db_to_cq_ctx* emu_db_to_cq_ctx);

/**
 * @brief Create emulated device msix.
 *
 * This function creates emulated device msix infrastructure.
 *
 * @param[in] process - A pointer to the emulation manager Flex IO process.
 * @param[in] target_ibv_ctx - A pointer to a SF device context.
 * @param[in] direct_mode - a flag to indicate EQ is to be used directly by app.
 * @param[in] target_uar_id - uar page id to be used for emulated cq created, relevant when
 *                            direct_mode is false.
 * @param[in] msix_id - The msix id.
 * @param[in] map - A pointer to msix  to eq mapping.
 * @param[out] msix - A pointer to receive the flexio_msix handle.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_create(struct flexio_process *process,
						 struct ibv_context *target_ibv_ctx,
						 uint8_t direct_mode, uint32_t target_uar_id,
						 uint16_t msix_id,
						 struct flexio_emu_device_msix_map* map,
						 struct flexio_msix** msix);

/**
 * @brief destroy emulated device msix.
 *
 * This function destroy emulated device msix infrastructure.
 *
 * @param[in] msix - A pointer to the flexio_msix handle structure.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_destroy(struct flexio_msix* msix);

/**
 * @brief Creates a Flex IO QP.
 *
 * This function creates a Flex IO QP.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * 			 If NULL - process' will be used.
 * @param[in]  qp_fattr - A pointer to the QP attributes struct.
 * @param[out] qp_ptr - A pointer to the created QP context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       struct flexio_qp_attr *qp_fattr, struct flexio_qp **qp_ptr);

/**
 * @brief Modify Flex IO QP.
 *
 * This function modifies Flex IO QP and transition it between states.
 * At the end of the procedure Flex IO QP would have moved from it's current state to to next state,
 * given in the fattr, if the move is a legal transition in the QP's state machine.
 *
 * @param[in] qp - A pointer to the QP context.
 * @param[in] fattr - A pointer to the QP attributes struct that will also define the QP connection.
 * @param[in] mask - A pointer to the optional QP attributes mask.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_modify(struct flexio_qp *qp, struct flexio_qp_attr *fattr,
			       struct flexio_qp_attr_opt_param_mask *mask);

/**
 * @brief Destroys a Flex IO QP.
 *
 * This function destroys a Flex IO QP.
 *
 * @param[in] qp - A pointer to the QP context.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_destroy(struct flexio_qp *qp);

/**
 * @brief Get a list of FlexIO Apps that are available.
 *
 * This function returns a list of Flex IO apps that are loaded.
 *
 * @param[out] app_list - A list of apps that are available.
 * @param[in/out] num_apps - number of apps to obtain / obtained.
 *
 * @return flexio status value.
 */
flexio_status flexio_app_get_list(struct flexio_app ***app_list, uint32_t *num_apps);

/**
 * @brief Free the list of flexio apps.
 *
 * This function frees the list of apps obtained from `flexio_app_get_list`.
 *
 * @param[in] apps - list obtained previously.
 *
 * @return flexio status value.
 */
flexio_status flexio_app_list_free(struct flexio_app **apps_list);

#define FLEXIO_MAX_NAME_LEN	(256)
/**
 * @brief Create a container for a FlexIO App.
 *
 * This function creates a named app with a given ELF. The ELF pointer
 * provided is copied into the library.
 * It is called from within the constructor generated by the compiler.
 *
 * @param[in] name - Name to assign the app (used in discovery later).
 *                   length of name should be up to FLEXIO_MAX_NAME_LEN bytes.
 * @param[in] elf - pointer to the ELF.
 * @param[in] elf_size - size of the elf.
 * @param[in/out] app - created app.
 *
 * @return flexio status value.
 */
flexio_status flexio_app_create(const char *name, void *elf, size_t elf_size,
				struct flexio_app **app);

/**
 * @brief Destroy a flexio app.
 *
 * This function destroys the state associated with the app and all registered functions.
 * This function will free the internal elf buffer.
 * It is called from within the destructor generated by the compiler.
 *
 * @param[in] app - App that was created before.
 *
 * @return flexio status value.
 */
flexio_status flexio_app_destroy(struct flexio_app *app);

/**
 * @brief  Callback function to pack the arguments for a function.
 *
 * This function is called internally from the FlexIO runtime upon user
 * making a call (e.g., flexio_process_call).
 * It packs the arguments for a user function into the argument buffer provided in `argbuf`.
 * The argument list can be arbitrarily long and is represented by `ap`. The correct usage
 * of this function requires the caller to initialize the list using `va_start`.
 *
 * @param argbuf [in] - Argument buffer of appropriate size to pack into arguments.
 * @param ap [in] - Variable argument list.
 *
 * @return flexio status value.
 */
typedef void (flexio_func_arg_pack_fn_t) (void *argbuf, va_list ap);

/**
 * @brief Register a function name at application start.
 *
 * This function registers the function name, stub address with the runtime.
 * It is called from within the constructor generated by the compiler.
 * @param[in] app - App that created before.
 * @param[in] dev_func_name - The device function name (entry point).
 * @param[in] dev_unpack_func_name - The device wrapper function that unpacks the argument buffer.
 * @param[in] host_stub_func_addr - The host stub function that is used by the application to
 *                                  reference the device function.
 * @param[in] argbuf_size - Size of the argument buffer required by this function.
 * @param[in] host_pack_func - Host callback function that packs the arguments.
 *
 * @return flexio status value.
 */
flexio_status flexio_func_pup_register(struct flexio_app *app,
				       const char *dev_func_name, const char *dev_unpack_func_name,
				       flexio_func_t *host_stub_func_addr, size_t argbuf_size,
				       flexio_func_arg_pack_fn_t *host_pack_func);

/**
 * @brief Register a function to be used later.
 *
 * This function is intended to be called directly by user in the situation where they
 * don’t desire pack/unpack support that is typically done by the compiler interface.
 *
 * It is the user’s responsibility to ensure that a function that was registered for RPC
 * has the type: flexio_dev_rpc_handler_t and the function registered for event handler
 * is: flexio_dev_event_handler_t.
 * The runtime will not provide any type checking. A mismatched call, such as using the
 * out_func of an Event handler for flexio_process_call() will result in undefined behavior.
 *
 * @param[in] app - previously created flexio app.
 * @param[in] dev_func_name - name of flexio function on device that will be called.
 * @param[out] out_func - opaque handle to use with flexio_process_call(),
 *                        flexio_event_handler_create(), …
 *
 * @return flexio status value.
 */
flexio_status flexio_func_register(struct flexio_app *app, const char *dev_func_name,
				   flexio_func_t **out_func);

/**
 * @brief Obtain info for previously registered function.
 *
 * This function is used to obtain info about a previously registered function.
 * It is used to compose higher-level libraries on top of DPACC / FlexIO interface.
 * It is not intended to be used directly by the user.
 *
 * The caller must ensure that the string pointers have been allocated
 * and are at least `FLEXIO_MAX_NAME_LEN` long to ensure that the call
 * doesn’t fail due to insufficient space for the function name.
 *
 * @param[in] app - FlexIO app.
 * @param[in] host_stub_func_addr - Known host stub func addr.
 * @param[out] pup - Whether function has been registered with pack/unpack support (0: No, 1:Yes).
 * @param[out] dev_func_name - Name of device function.
 * @param[out] dev_unpack_func_name - Name of unpack routine on device, NA if pup==0.
 * @param[in] func_name_len - Size of function name len allocated.
 * @param[out] argbuf_size - Size of argument buffer, NA if pup==0.
 * @param[out] host_pack_func - Function pointer to host packing routine, NA if pup==0.
 * @param[out] dev_func_addr - address of device function.
 * @param[out] dev_unpack_func_addr - address of device unpack function.
 *
 * @return flexio status value.
*/
flexio_status flexio_func_get_register_info(struct flexio_app *app,
					    flexio_func_t *host_stub_func_addr, uint32_t *pup,
					    char *dev_func_name, char *dev_unpack_func_name,
					    size_t func_name_len, size_t *argbuf_size,
					    flexio_func_arg_pack_fn_t **host_pack_func,
					    flexio_uintptr_t *dev_func_addr,
					    flexio_uintptr_t *dev_unpack_func_addr);

/**
 * @brief Retrieve ELF binary associated with application.
 *
 * This function registers the function name, stub address with the runtime.
 * Compiler calls this from within the constructor.
 * @param[in] app - App that created before.
 * @param[in] bin_buff - Pointer to buffer to copy ELF binary.
 * @param[in] bin_size - Size of buffer pointed by bin_buff.
 *                       If parameter is smaller than ELF binary size function will fail.
 *
 * @return flexio status value.
 */
flexio_status flexio_app_get_elf(struct flexio_app *app, uint64_t *bin_buff, size_t bin_size);

/* Helpers and Getters */
struct flexio_thread *flexio_event_handler_get_thread(struct flexio_event_handler *event_handler);
uint32_t flexio_event_handler_get_id(struct flexio_event_handler *event_handler);
uint32_t flexio_cq_get_cq_num(struct flexio_cq *cq);
uint32_t flexio_rq_get_wq_num(struct flexio_rq *rq);
struct mlx5dv_devx_obj *flexio_rq_get_tir(struct flexio_rq *rq);
uint32_t flexio_sq_get_wq_num(struct flexio_sq *sq);
uint32_t flexio_qp_get_qp_num(struct flexio_qp *qp);
struct mlx5dv_devx_uar *flexio_uar_get_devx_uar(struct flexio_uar *flexio_uar);
struct flexio_uar *flexio_outbox_get_uar(struct flexio_outbox *outbox);
uint32_t flexio_outbox_get_id(struct flexio_outbox *outbox);
uint32_t flexio_window_get_id(struct flexio_window *window);
uint32_t flexio_emu_db_to_cq_ctx_get_id(struct flexio_emu_db_to_cq_ctx *emu_db_to_cq_ctx);
uint32_t flexio_emulated_device_msix_get_eqn(struct flexio_msix *msix);
uint32_t flexio_emulated_device_msix_get_cqn(struct flexio_msix *msix);
uint32_t flexio_mkey_get_id(struct flexio_mkey *mkey);
const char* flexio_app_get_name(struct flexio_app *app);
size_t flexio_app_get_elf_size(struct flexio_app *app);

#ifdef __cplusplus
}
#endif
#endif /* _FLEXIO_SDK_H_ */
